<?php

/**
 * @file
 * Contains \Drupal\entityqueryapi\Controller\EntityQueryApiController.
 */

namespace Drupal\entityqueryapi\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\Query\QueryFactory;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class EntityQueryApiController.
 *
 * @package Drupal\entityqueryapi\Controller
 */
class EntityQueryApiController extends ControllerBase {

  protected $entityTypeManager;
  protected $entityType;
  protected $query_factory;
  protected $normalizer;

  public function __construct(EntityTypeManagerInterface $entity_type_manager, QueryFactory $query_factory, NormalizerInterface $normalizer, SerializerInterface $serializer) {
    $normalizer->setSerializer($serializer);

    $this->entityTypeManager = $entity_type_manager;
    $this->queryFactory = $query_factory;
    $this->normalizer = $normalizer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $query_factory = $container->get('entity.query');
    return new static(
      $container->get('entity_type.manager'),
      $query_factory,
      $container->get('serializer.normalizer.list'),
      $container->get('serializer')
    );
  }

  /**
   * List activity_actions.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Return a JSON Response
   */
  public function content(Request $request, EntityTypeInterface $entity_type) {
    $this->entityType = $entity_type;

    $options =  $this->getQueryOptions($this->getQueryParams($request));
    $entities = $this->getEntities($options);

    $normalized = $this->normalizer->normalize($entities, 'json');

    $response = new JsonResponse();
    $response->setData($normalized);
    return $response;
  }

  protected function getEntities($options) {
    $entity_type_id = $this->getEntityTypeId();

    $query = $this->queryFactory->get($entity_type_id);

    // For every filter, create an entityQuery condition.
    if (!empty($options['filters'])) {
      $query = array_reduce($options['filters'], function ($query, $filter) {
        return $query->condition(
          $filter['field'],
          isset($filter['value']) ? $filter['value'] : NULL,
          isset($filter['operator']) ? $filter['operator'] : NULL,
          isset($filter['langcode']) ? $filter['langcode'] : NULL
        );
      }, $query);
    }

    // For every sort, create an entityQuery sort.
    if (!empty($options['sorts'])) {
      $query = array_reduce($options['sorts'], function ($query, $sort) {
        return $query->sort(
          $sort['field'],
          isset($sort['direction']) ? $sort['direction'] : 'ASC',
          isset($sort['langcode']) ? $sort['langcode'] : NULL
        );
      }, $query);
    }

    // If range options are defined, apply them to the query.
    if (isset($options['range'])) {
      $start = (isset($options['range']['start'])) ? $options['start'] : NULL;
      $length = (isset($options['range']['length'])) ? $options['length'] : NULL;
      $query->range($start, $length);
    }

    // Execute and get the entity IDs from the result.
    $entity_ids = array_values($query->execute());

    // Get a storage connection for the appropriate entity type.
    $storage = $this->entityTypeManager()->getStorage($entity_type_id);

    // Load all the entities from their ids.
    return $storage->loadMultiple($entity_ids);
  }

  protected function getQueryParams($request) {
    return $request->query;
  }

  protected function getQueryOptions($params) {
    $option_keys = $this->getValidOptionKeys($params->keys());
    $options['filters'] = $this->getQueryFilters($option_keys['filters'], $params);
    $options['sorts'] = $this->getQuerySorts($option_keys['sorts'], $params);
    if ($option_keys['range']) $options['range'] = $params->get($option_keys['range']);

    return $options;
  }

  protected static function getQuerySorts($keys, $params) {
    $get_sort_param_func = function ($params) {
      return function ($key) use ($params) {
        return $params->get($key);
      };
    };

    $validate_sort_param = function ($sort_param) {
      $validate_sort_direction = function ($sort_direction) {
        return in_array($sort_direction, array(
          'ASC',
          'DESC',
        ));
      };

      return (
        isset($sort_param['field']) &&
        (!isset($sort_param['direction']) || $validate_sort_direction($sort_param['direction']))
      );
    };

    $sorts = array_filter(
      array_map($get_sort_param_func($params), $keys),
      $validate_sort_param
    );

    return $sorts;
  }

  protected static function getQueryFilters($keys, $params) {
    $get_filter_param_func = function ($params) {
      return function ($key) use ($params) {
        $operator_dict = array(
          'EQ' => '=',
          'NOT_EQ' => '<>',
          'GT' => '>',
          'GTEQ' => '>=',
          'LT' => '<',
          'LTEQ' => '<=',
        );

        $filter = $params->get($key);
        if (isset($filter['operator']) && key_exists($filter['operator'], $operator_dict)) {
          $filter['operator'] = $operator_dict[$filter['operator']];
        }

        return $filter;
      };
    };

    $validate_filter_param = function ($filter_param) {
      $validate_filter_op = function ($filter_op) {
        return in_array($filter_op, array(
          '=',
          '<>',
          '>',
          '>=',
          '<',
          '<=',
          'STARTS_WITH',
          'CONTAINS',
          'ENDS_WITH',
          'IN',
          'NOT_IN',
          'BETWEEN',
        ));
      };

      return (
        isset($filter_param['field']) &&
        isset($filter_param['value']) &&
        $validate_filter_op($filter_param['operator'])
      );
    };

    $filters = array_filter(
      array_map($get_filter_param_func($params), $keys),
      $validate_filter_param
    );

    return $filters;
  }

  protected static function getValidOptionKeys($keys) {
    $reduce_keys = function ($option_keys, $key) {
      if (preg_match('/^filter_\d+$/', $key)) $option_keys['filters'][] = $key;
      if (preg_match('/^sort_\d+$/', $key))   $option_keys['sorts'][] = $key;
      if (preg_match('/^range$/', $key))      $option_keys['range'] = $key;
      return $option_keys;
    };

    $option_keys = array_reduce($keys, $reduce_keys, array(
      'filters' => array(),
      'sorts' => array(),
      'range' => null,
    ));

    return $option_keys;
  }

  protected function getEntityTypeId() {
    return $this->entityType->id();
  }

}
